package org.warp.cowdb;

import org.warp.cowdb.database.*;

import java.io.IOException;
import java.nio.file.Files;
import java.nio.file.Path;

public class Database implements IDatabase, IDatabaseTools {

	private final IDatabaseTools databaseTools;
	private final DatabaseFileIO fileIO;
	private final DatabaseBlocksIO blocksIO;
	private final DatabaseBlocksMetadata blocksMetadata;
	private final DatabaseReferencesIO referencesIO;
	private final DatabaseReferencesMetadata referencesMetadata;
	private final DatabaseObjectsIO objectsIO;
	private EnhancedObject loadedRootObject;

	public Database(Path dataFile, Path blocksMetaFile, Path referencesMetaFile) throws IOException {
		if (Files.notExists(dataFile)) {
			Files.createFile(dataFile);
		}
		if (Files.notExists(blocksMetaFile)) {
			Files.createFile(blocksMetaFile);
		}
		if (Files.notExists(referencesMetaFile)) {
			Files.createFile(referencesMetaFile);
		}
		this.databaseTools = this;
		this.fileIO = new DatabaseFileIO(dataFile);
		this.blocksMetadata = new DatabaseBlocksMetadata(blocksMetaFile);
		this.blocksIO = new DatabaseBlocksIO(fileIO, blocksMetadata);
		this.referencesMetadata = new DatabaseReferencesMetadata(referencesMetaFile);
		this.referencesIO = new DatabaseReferencesIO(blocksIO, referencesMetadata);
		this.objectsIO = new DatabaseObjectsIO(databaseTools, referencesIO);
	}

	@Override
	public void close() throws IOException {
		this.objectsIO.setEnhancedObject(0, loadedRootObject);
		this.referencesMetadata.close();
		this.blocksMetadata.close();
		this.fileIO.close();
	}

	public <T extends EnhancedObject> T loadRoot(Class<T> type, FunctionWithIO<IDatabaseTools, T> ifAbsent) throws IOException {
		if (loadedRootObject != null) {
			throw new RuntimeException("Root already set!");
		}
		T root;
		if (referencesMetadata.getFirstFreeReference() > 0) {
			root = objectsIO.loadEnhancedObject(0, type);
		} else {
			if (objectsIO.newNullObject() != 0) {
				throw new IOException("Can't allocate root!");
			} else {
				root = ifAbsent.apply(Database.this);
				objectsIO.setEnhancedObject(0, root);
			}
		}
		loadedRootObject = root;
		return root;
	}

	protected void registerClass(Class<?> type, int id) {
		this.objectsIO.registerClass(type, id);
	}

	@Override
	public void initializeEnhancedObject(EnhancedObject enhancedObject) throws IOException {
		this.objectsIO.getDataInitializer().initializeDBObject(enhancedObject);
	}

	@Override
	public IObjectsIO getObjectsIO() {
		return objectsIO;
	}

}
